@@ -3,7 +3,7 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2002 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2003 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -31,9 +31,39 @@
 #undef HUGE
 #undef CHARBITS
 #undef INTBITS
+#if HAVE_INTTYPES_H
+# include <inttypes.h>
+#else
+# if HAVE_STDINT_H
+#  include <stdint.h>
+# endif
+#endif
 #include <math.h>
 #include "random.h"
 
+#ifndef CHAR_BIT
+# define CHAR_BIT 8
+#endif
+
+/* The extra casts work around common compiler bugs.  */
+#define TYPE_SIGNED(t) (! ((t) 0 < (t) -1))
+/* The outer cast is needed to work around a bug in Cray C 5.0.3.0.
+   It is necessary at least when t == time_t.  */
+#define TYPE_MINIMUM(t) ((t) (TYPE_SIGNED (t) \
+			      ? ~ (t) 0 << (sizeof (t) * CHAR_BIT - 1) : (t) 0))
+#define TYPE_MAXIMUM(t) ((t) (~ (t) 0 - TYPE_MINIMUM (t)))
+
+#ifndef INTMAX_MIN
+# define INTMAX_MIN TYPE_MINIMUM (intmax_t)
+#endif
+#ifndef UINTMAX_MAX
+# define UINTMAX_MAX TYPE_MAXIMUM (uintmax_t)
+#endif
+
+#ifndef SIZE_MAX	/* C99 constant, can't rely on it everywhere */
+#define SIZE_MAX ((size_t) -1)
+#endif
+
 /* can declare these, since we always use the random shipped with gawk */
 extern char *initstate P((unsigned long seed, char *state, long n));
 extern char *setstate P((char *state));
@@ -43,7 +73,7 @@ extern void srandom P((unsigned long seed));
 extern NODE **fields_arr;
 extern int output_is_tty;
 
-static NODE *sub_common P((NODE *tree, int how_many, int backdigs));
+static NODE *sub_common P((NODE *tree, long how_many, int backdigs));
 
 #ifdef _CRAY
 /* Work around a problem in conversion of doubles to exact integers. */
@@ -118,7 +148,7 @@ do_exp(NODE *tree)
 	double d, res;
 
 	tmp = tree_eval(tree->lnode);
-	if (do_lint && (tmp->flags & (NUM|NUMBER)) == 0)
+	if (do_lint && (tmp->flags & (NUMCUR|NUMBER)) == 0)
 		lintwarn(_("exp: received non-numeric argument"));
 	d = force_number(tmp);
 	free_temp(tmp);
@@ -138,7 +168,7 @@ do_exp(NODE *tree)
  */
 
 static FILE *
-stdfile(char *name, size_t len)
+stdfile(const char *name, size_t len)
 {
 	if (len == 11) {
 		if (STREQN(name, "/dev/stderr", 11))
@@ -159,7 +189,7 @@ do_fflush(NODE *tree)
 	NODE *tmp;
 	FILE *fp;
 	int status = 0;
-	char *file;
+	const char *file;
 
 	/* fflush() --- flush stdout */
 	if (tree == NULL) {
@@ -275,13 +305,13 @@ NODE *
 do_index(NODE *tree)
 {
 	NODE *s1, *s2;
-	register char *p1, *p2;
+	register const char *p1, *p2;
 	register size_t l1, l2;
 	long ret;
 #ifdef MBS_SUPPORT
 	size_t mbclen = 0;
 	mbstate_t mbs1, mbs2;
-	if (MB_CUR_MAX > 1) {
+	if (gawk_mb_cur_max > 1) {
 		memset(&mbs1, 0, sizeof(mbstate_t));
 		memset(&mbs2, 0, sizeof(mbstate_t));
 	}
@@ -291,9 +321,9 @@ do_index(NODE *tree)
 	s1 = tree_eval(tree->lnode);
 	s2 = tree_eval(tree->rnode->lnode);
 	if (do_lint) {
-		if ((s1->flags & (STRING|STR)) == 0)
+		if ((s1->flags & (STRING|STRCUR)) == 0)
 			lintwarn(_("index: received non-string first argument"));
-		if ((s2->flags & (STRING|STR)) == 0)
+		if ((s2->flags & (STRING|STRCUR)) == 0)
 			lintwarn(_("index: received non-string second argument"));
 	}
 	force_string(s1);
@@ -320,7 +350,7 @@ do_index(NODE *tree)
 			if (l2 > l1)
 				break;
 #ifdef MBS_SUPPORT
-			if (MB_CUR_MAX > 1) {
+			if (gawk_mb_cur_max > 1) {
 				if (strncasecmpmbs(p1, mbs1, p2, mbs2, l2) == 0) {
 					ret = 1 + s1->stlen - l1;
 					break;
@@ -357,7 +387,7 @@ do_index(NODE *tree)
 				break;
 			}
 #ifdef MBS_SUPPORT
-			if (MB_CUR_MAX > 1) {
+			if (gawk_mb_cur_max > 1) {
 				mbclen = mbrlen(p1, l1, &mbs1);
 				if ((mbclen == 1) || (mbclen == (size_t) -1) ||
 					(mbclen == (size_t) -2) || (mbclen == 0)) {
@@ -403,7 +433,7 @@ do_int(NODE *tree)
 	double d;
 
 	tmp = tree_eval(tree->lnode);
-	if (do_lint && (tmp->flags & (NUM|NUMBER)) == 0)
+	if (do_lint && (tmp->flags & (NUMCUR|NUMBER)) == 0)
 		lintwarn(_("int: received non-numeric argument"));
 	d = force_number(tmp);
 	d = double_to_int(d);
@@ -420,7 +450,7 @@ do_length(NODE *tree)
 	size_t len;
 
 	tmp = tree_eval(tree->lnode);
-	if (do_lint && (tmp->flags & (STRING|STR)) == 0)
+	if (do_lint && (tmp->flags & (STRING|STRCUR)) == 0)
 		lintwarn(_("length: received non-string argument"));
 	len = force_string(tmp)->stlen;
 	free_temp(tmp);
@@ -436,7 +466,7 @@ do_log(NODE *tree)
 	double d, arg;
 
 	tmp = tree_eval(tree->lnode);
-	if (do_lint && (tmp->flags & (NUM|NUMBER)) == 0)
+	if (do_lint && (tmp->flags & (NUMCUR|NUMBER)) == 0)
 		lintwarn(_("log: received non-numeric argument"));
 	arg = (double) force_number(tmp);
 	if (arg < 0.0)
@@ -459,9 +489,9 @@ do_log(NODE *tree)
 NODE *
 format_tree(
 	const char *fmt_string,
-	int n0,
+	size_t n0,
 	register NODE *carg,
-	int num_args)
+	long num_args)
 {
 /* copy 'l' bytes from 's' to 'obufout' checking for space in the process */
 /* difference of pointers should be of ptrdiff_t type, but let us be kind */
@@ -513,7 +543,7 @@ format_tree(
 	int toofew = FALSE;
 	char *obuf, *obufout;
 	size_t osiz, ofre;
-	char *chbuf;
+	const char *chbuf;
 	const char *s0, *s1;
 	int cs1;
 	NODE *arg;
@@ -524,21 +554,21 @@ format_tree(
 #ifdef sun386		/* Can't cast unsigned (int/long) from ptr->value */
 	long tmp_uval;	/* on 386i 4.0.1 C compiler -- it just hangs */
 #endif
-	unsigned long uval;
+	uintmax_t uval;
 	int sgn;
 	int base = 0;
 	char cpbuf[30];		/* if we have numbers bigger than 30 */
 	char *cend = &cpbuf[30];/* chars, we lose, but seems unlikely */
 	char *cp;
-	char *fill;
+	const char *fill;
 	double tmpval;
 	char signchar = FALSE;
 	size_t len;
 	int zero_flag = FALSE;
-	static char sp[] = " ";
-	static char zero_string[] = "0";
-	static char lchbuf[] = "0123456789abcdef";
-	static char Uchbuf[] = "0123456789ABCDEF";
+	static const char sp[] = " ";
+	static const char zero_string[] = "0";
+	static const char lchbuf[] = "0123456789abcdef";
+	static const char Uchbuf[] = "0123456789ABCDEF";
 
 #define INITIAL_OUT_SIZE	512
 	emalloc(obuf, char *, INITIAL_OUT_SIZE, "format_tree");
@@ -643,7 +673,7 @@ format_tree(
 		s1++;
 
 retry:
-		if (n0-- <= 0)	/* ran out early! */
+		if (n0-- == 0)	/* ran out early! */
 			break;
 
 		switch (cs1 = *s1++) {
@@ -654,6 +684,13 @@ check_pos:
 			goto retry;
 		case '%':
 			need_format = FALSE;
+			/*
+			 * 29 Oct. 2002:
+			 * The C99 standard pages 274 and 279 seem to imply that
+			 * since there's no arg converted, the field width doesn't
+			 * apply.  The code already was that way, but this
+			 * comment documents it, at least in the code.
+			 */
 			bchunk_one("%");
 			s0 = s1;
 			break;
@@ -708,7 +745,7 @@ check_pos:
 				if (argnum <= 0)
 					fatal(_("arg count with `$' must be > 0"));
 				if (argnum >= num_args)
-					fatal(_("arg count %d greater than total number of supplied arguments"), argnum);
+					fatal(_("arg count %ld greater than total number of supplied arguments"), argnum);
 			} else
 				fatal(_("`$' not permitted after period in format"));
 			goto retry;
@@ -832,17 +869,20 @@ check_pos:
 				tmp_uval = arg->numbr; 
 				uval = (unsigned long) tmp_uval;
 #else
-				uval = (unsigned long) arg->numbr;
+				uval = (uintmax_t) arg->numbr;
 #endif
 				cpbuf[0] = uval;
 				prec = 1;
 				cp = cpbuf;
 				goto pr_tail;
 			}
-			if (have_prec == FALSE)
-				prec = 1;
-			else if (prec > arg->stlen)
-				prec = arg->stlen;
+			/*
+			 * As per POSIX, only output first character of a
+			 * string value.  Thus, we ignore any provided
+			 * precision, forcing it to 1.  (Didn't this
+			 * used to work? 6/2003.)
+			 */
+			prec = 1;
 			cp = arg->stptr;
 			goto pr_tail;
 		case 's':
@@ -869,17 +909,16 @@ check_pos:
 				goto pr_tail;
 
 			if (tmpval < 0) {
-				if (tmpval < LONG_MIN)
+				if (tmpval < INTMAX_MIN)
 					goto out_of_range;
 				sgn = TRUE;
-				uval = - (unsigned long) (long) tmpval;
+				uval = - (uintmax_t) (intmax_t) tmpval;
 			} else {
-				/* Use !, so that NaNs are out of range.
-				   The cast avoids a SunOS 4.1.x cc bug.  */
-				if (! (tmpval <= (unsigned long) ULONG_MAX))
+				/* Use !, so that NaNs are out of range.  */
+				if (! (tmpval <= UINTMAX_MAX))
 					goto out_of_range;
 				sgn = FALSE;
-				uval = (unsigned long) tmpval;
+				uval = (uintmax_t) tmpval;
 			}
 			do {
 				*--cp = (char) ('0' + uval % 10);
@@ -947,15 +986,14 @@ check_pos:
 				goto pr_tail;
 
 			if (tmpval < 0) {
-				if (tmpval < LONG_MIN)
+				if (tmpval < INTMAX_MIN)
 					goto out_of_range;
-				uval = (unsigned long) (long) tmpval;
+				uval = (uintmax_t) (intmax_t) tmpval;
 			} else {
-				/* Use !, so that NaNs are out of range.
-				   The cast avoids a SunOS 4.1.x cc bug.  */
-				if (! (tmpval <= (unsigned long) ULONG_MAX))
+				/* Use !, so that NaNs are out of range.  */
+				if (! (tmpval <= UINTMAX_MAX))
 					goto out_of_range;
-				uval = (unsigned long) tmpval;
+				uval = (uintmax_t) tmpval;
 			}
 			/*
 			 * When to fill with zeroes is of course not simple.
@@ -1013,6 +1051,9 @@ check_pos:
 
      out_of_range:
 			/* out of range - emergency use of %g format */
+			if (do_lint)
+				lintwarn(_("[s]printf: value %g is out of range for `%%%c' format"),
+							tmpval, cs1);
 			cs1 = 'g';
 			goto format_float;
 
@@ -1065,7 +1106,7 @@ check_pos:
 		if (toofew)
 			fatal("%s\n\t`%s'\n\t%*s%s",
 			      _("not enough arguments to satisfy format string"),
-			      fmt_string, s1 - fmt_string - 1, "",
+			      fmt_string, (int) (s1 - fmt_string - 1), "",
 			      _("^ ran out for this one"));
 	}
 	if (do_lint) {
@@ -1106,6 +1147,29 @@ do_sprintf(NODE *tree)
 	return r;
 }
 
+/*
+ * redirect_to_fp --- return fp for redirection, NULL on failure
+ * or stdout if no redirection, used by all print routines
+ */
+
+static inline FILE *
+redirect_to_fp(NODE *tree, struct redirect **rpp)
+{
+	int errflg;	/* not used, sigh */
+	struct redirect *rp;
+
+	if (tree == NULL)
+		return stdout;
+
+	rp = redirect(tree, &errflg);
+	if (rp != NULL) {
+		*rpp = rp;
+		return  rp->fp;
+	}
+
+	return NULL;
+}
+
 /* do_printf --- perform printf, including redirection */
 
 void
@@ -1123,18 +1187,9 @@ do_printf(NODE *tree)
 		fatal(_("printf: no arguments"));
 	}
 
-	if (tree->rnode != NULL) {
-		int errflg;	/* not used, sigh */
-
-		rp = redirect(tree->rnode, &errflg);
-		if (rp != NULL) {
-			fp = rp->fp;
-			if (fp == NULL)
-				return;
-		} else
-			return;
-	} else
-		fp = stdout;
+	fp = redirect_to_fp(tree->rnode, & rp);
+	if (fp == NULL)
+		return;
 	tree->lnode->printf_count = tree->printf_count;
 	tree = do_sprintf(tree->lnode);
 	efwrite(tree->stptr, sizeof(char), tree->stlen, fp, "printf", rp, TRUE);
@@ -1152,7 +1207,7 @@ do_sqrt(NODE *tree)
 	double arg;
 
 	tmp = tree_eval(tree->lnode);
-	if (do_lint && (tmp->flags & (NUM|NUMBER)) == 0)
+	if (do_lint && (tmp->flags & (NUMCUR|NUMBER)) == 0)
 		lintwarn(_("sqrt: received non-numeric argument"));
 	arg = (double) force_number(tmp);
 	free_temp(tmp);
@@ -1177,7 +1232,8 @@ do_substr(NODE *tree)
 	d_index = force_number(t2);
 	free_temp(t2);
 
-	if (d_index < 1.0) {
+	/* the weird `! (foo)' tests help catch NaN values. */
+	if (! (d_index >= 1)) {
 		if (do_lint)
 			lintwarn(_("substr: start index %g is invalid, using 1"),
 				 d_index);
@@ -1187,48 +1243,66 @@ do_substr(NODE *tree)
 		lintwarn(_("substr: non-integer start index %g will be truncated"),
 			 d_index);
 
-	indx = d_index - 1;	/* awk indices are from 1, C's are from 0 */
+	/* awk indices are from 1, C's are from 0 */
+	if (d_index <= SIZE_MAX)
+		indx = d_index - 1;
+	else
+		indx = SIZE_MAX;
 
 	if (tree->rnode->rnode == NULL) {	/* third arg. missing */
 		/* use remainder of string */
 		length = t1->stlen - indx;
+		d_length = length;	/* set here in case used in diagnostics, below */
 	} else {
 		t3 = tree_eval(tree->rnode->rnode->lnode);
 		d_length = force_number(t3);
 		free_temp(t3);
-		if (d_length <= 0.0) {
-			if (do_lint)
-				lintwarn(_("substr: length %g is <= 0"), d_length);
+		if (! (d_length >= 1)) {
+			if (do_lint == LINT_ALL)
+				lintwarn(_("substr: length %g is not >= 1"), d_length);
+			else if (do_lint == LINT_INVALID && ! (d_length >= 0))
+				lintwarn(_("substr: length %g is not >= 0"), d_length);
 			free_temp(t1);
 			return Nnull_string;
 		}
-		if (do_lint && double_to_int(d_length) != d_length)
-			lintwarn(
-		_("substr: non-integer length %g will be truncated"),
-				d_length);
-		length = d_length;
+		if (do_lint) {
+			if (double_to_int(d_length) != d_length)
+				lintwarn(
+			_("substr: non-integer length %g will be truncated"),
+					d_length);
+
+			if (d_length > SIZE_MAX)
+				lintwarn(
+			_("substr: length %g too big for string indexing, truncating to %g"),
+					d_length, (double) SIZE_MAX);
+		}
+		if (d_length < SIZE_MAX)
+			length = d_length;
+		else
+			length = SIZE_MAX;
 	}
 
 	if (t1->stlen == 0) {
-		if (do_lint)
+		/* substr("", 1, 0) produces a warning only if LINT_ALL */
+		if (do_lint && (do_lint == LINT_ALL || ((indx | length) != 0)))
 			lintwarn(_("substr: source string is zero length"));
 		free_temp(t1);
 		return Nnull_string;
 	}
-	if ((indx + length) > t1->stlen) {
-		if (do_lint)
-			lintwarn(
-	_("substr: length %d at start index %d exceeds length of first argument (%d)"),
-			length, indx+1, t1->stlen);
-		length = t1->stlen - indx;
-	}
 	if (indx >= t1->stlen) {
 		if (do_lint)
-			lintwarn(_("substr: start index %d is past end of string"),
-				indx+1);
+			lintwarn(_("substr: start index %g is past end of string"),
+				d_index);
 		free_temp(t1);
 		return Nnull_string;
 	}
+	if (length > t1->stlen - indx) {
+		if (do_lint)
+			lintwarn(
+	_("substr: length %g at start index %g exceeds length of first argument (%lu)"),
+			d_length, d_index, (unsigned long int) t1->stlen);
+		length = t1->stlen - indx;
+	}
 	r = tmp_string(t1->stptr + indx, length);
 	free_temp(t1);
 	return r;
@@ -1246,8 +1320,8 @@ do_strftime(NODE *tree)
 	size_t buflen, bufsize;
 	char buf[BUFSIZ];
 	/* FIXME: One day make %d be %e, after C 99 is common. */
-	static char def_format[] = "%a %b %d %H:%M:%S %Z %Y";
-	char *format;
+	static const char def_format[] = "%a %b %d %H:%M:%S %Z %Y";
+	const char *format;
 	int formatlen;
 
 	/* set defaults first */
@@ -1259,7 +1333,7 @@ do_strftime(NODE *tree)
 	if (tree != NULL) {	/* have args */
 		if (tree->lnode != NULL) {
 			NODE *tmp = tree_eval(tree->lnode);
-			if (do_lint && (tmp->flags & (STRING|STR)) == 0)
+			if (do_lint && (tmp->flags & (STRING|STRCUR)) == 0)
 				lintwarn(_("strftime: received non-string first argument"));
 			t1 = force_string(tmp);
 			format = t1->stptr;
@@ -1274,7 +1348,7 @@ do_strftime(NODE *tree)
 	
 		if (tree->rnode != NULL) {
 			t2 = tree_eval(tree->rnode->lnode);
-			if (do_lint && (t2->flags & (NUM|NUMBER)) == 0)
+			if (do_lint && (t2->flags & (NUMCUR|NUMBER)) == 0)
 				lintwarn(_("strftime: received non-numeric second argument"));
 			fclock = (time_t) force_number(t2);
 			free_temp(t2);
@@ -1315,7 +1389,7 @@ do_strftime(NODE *tree)
 /* do_systime --- get the time of day */
 
 NODE *
-do_systime(NODE *tree)
+do_systime(NODE *tree ATTRIBUTE_UNUSED)
 {
 	time_t lclock;
 
@@ -1337,7 +1411,7 @@ do_mktime(NODE *tree)
 	char save;
 
 	t1 = tree_eval(tree->lnode);
-	if (do_lint && (t1->flags & (STRING|STR)) == 0)
+	if (do_lint && (t1->flags & (STRING|STRCUR)) == 0)
 		lintwarn(_("mktime: received non-string argument"));
 	t1 = force_string(t1);
 
@@ -1382,7 +1456,7 @@ do_system(NODE *tree)
 
 	(void) flush_io();     /* so output is synchronous with gawk's */
 	tmp = tree_eval(tree->lnode);
-	if (do_lint && (tmp->flags & (STRING|STR)) == 0)
+	if (do_lint && (tmp->flags & (STRING|STRCUR)) == 0)
 		lintwarn(_("system: received non-string argument"));
 	cmd = force_string(tmp)->stptr;
 
@@ -1408,7 +1482,8 @@ do_system(NODE *tree)
 
 		os_restore_mode(fileno(stdin));
 		ret = system(cmd);
-		ret = (ret >> 8) & 0xff;
+		if (ret != -1)
+			ret = WEXITSTATUS(ret);
 		if ((BINMODE & 1) != 0)
 			os_setbinmode(fileno(stdin), O_BINARY);
 
@@ -1432,18 +1507,9 @@ do_print(register NODE *tree)
 	NODE *save;
 	NODE *tval;
 
-	if (tree->rnode) {
-		int errflg;		/* not used, sigh */
-
-		rp = redirect(tree->rnode, &errflg);
-		if (rp != NULL) {
-			fp = rp->fp;
-			if (fp == NULL)
-				return;
-		} else
-			return;
-	} else
-		fp = stdout;
+	fp = redirect_to_fp(tree->rnode, & rp);
+	if (fp == NULL)
+		return;
 
 	/*
 	 * General idea is to evaluate all the expressions first and
@@ -1493,6 +1559,36 @@ do_print(register NODE *tree)
 	free(t);
 }
 
+/* do_print_rec --- special case printing of $0, for speed */
+
+void 
+do_print_rec(register NODE *tree)
+{
+	struct redirect *rp = NULL;
+	register FILE *fp;
+	register NODE *f0;
+
+	fp = redirect_to_fp(tree->rnode, & rp);
+	if (fp == NULL)
+		return;
+
+	if (! field0_valid)
+		(void) get_field(0L, NULL);	/* rebuild record */
+
+	f0 = fields_arr[0];
+
+	if (do_lint && f0 == Nnull_string)
+		lintwarn(_("reference to uninitialized field `$%d'"), 0);
+
+	efwrite(f0->stptr, sizeof(char), f0->stlen, fp, "print", rp, FALSE);
+
+	if (ORSlen > 0)
+		efwrite(ORS, sizeof(char), (size_t) ORSlen, fp, "print", rp, TRUE);
+
+	if (rp != NULL && (rp->flag & RED_TWOWAY) != 0)
+		fflush(rp->fp);
+}
+
 /* do_tolower --- lower case a string */
 
 NODE *
@@ -1503,37 +1599,37 @@ do_tolower(NODE *tree)
 #ifdef MBS_SUPPORT
 	size_t mbclen = 0;
 	mbstate_t mbs, prev_mbs;
-	if (MB_CUR_MAX > 1)
+	if (gawk_mb_cur_max > 1)
 		memset(&mbs, 0, sizeof(mbstate_t));
 #endif
 
 	t1 = tree_eval(tree->lnode);
-	if (do_lint && (t1->flags & (STRING|STR)) == 0)
+	if (do_lint && (t1->flags & (STRING|STRCUR)) == 0)
 		lintwarn(_("tolower: received non-string argument"));
 	t1 = force_string(t1);
 	t2 = tmp_string(t1->stptr, t1->stlen);
 	for (cp = (unsigned char *)t2->stptr,
 	     cp2 = (unsigned char *)(t2->stptr + t2->stlen); cp < cp2; cp++)
 #ifdef MBS_SUPPORT
-		if (MB_CUR_MAX > 1) {
+		if (gawk_mb_cur_max > 1) {
 			wchar_t wc;
 			prev_mbs = mbs;
-			mbclen = (size_t) mbrtowc(&wc, cp, cp2 - cp, &mbs);
+			mbclen = (size_t) mbrtowc(&wc, (char *) cp, cp2 - cp,
+						  &mbs);
 			if ((mbclen != 1) && (mbclen != (size_t) -1) &&
 				(mbclen != (size_t) -2) && (mbclen != 0)) {
 				/* a multibyte character.  */
-				if (iswupper(wc))
-				{
+				if (iswupper(wc)) {
 					wc = towlower(wc);
-					wcrtomb(cp, wc, &prev_mbs);
+					wcrtomb((char *) cp, wc, &prev_mbs);
 				}
 				/* Adjust the pointer.  */
 				cp += mbclen - 1;
-		    } else {
+			} else {
 				/* Otherwise we treat it as a singlebyte character.  */
 				if (ISUPPER(*cp))
 					*cp = tolower(*cp);
-		    }
+			}
 		} else
 #endif
 		if (ISUPPER(*cp))
@@ -1552,37 +1648,37 @@ do_toupper(NODE *tree)
 #ifdef MBS_SUPPORT
 	size_t mbclen = 0;
 	mbstate_t mbs, prev_mbs;
-	if (MB_CUR_MAX > 1)
+	if (gawk_mb_cur_max > 1)
 		memset(&mbs, 0, sizeof(mbstate_t));
 #endif
 
 	t1 = tree_eval(tree->lnode);
-	if (do_lint && (t1->flags & (STRING|STR)) == 0)
+	if (do_lint && (t1->flags & (STRING|STRCUR)) == 0)
 		lintwarn(_("toupper: received non-string argument"));
 	t1 = force_string(t1);
 	t2 = tmp_string(t1->stptr, t1->stlen);
 	for (cp = (unsigned char *)t2->stptr,
 	     cp2 = (unsigned char *)(t2->stptr + t2->stlen); cp < cp2; cp++)
 #ifdef MBS_SUPPORT
-		if (MB_CUR_MAX > 1) {
+		if (gawk_mb_cur_max > 1) {
 			wchar_t wc;
 			prev_mbs = mbs;
-			mbclen = (size_t) mbrtowc(&wc, cp, cp2 - cp, &mbs);
+			mbclen = (size_t) mbrtowc(&wc, (char *) cp, cp2 - cp,
+						  &mbs);
 			if ((mbclen != 1) && (mbclen != (size_t) -1) &&
 				(mbclen != (size_t) -2) && (mbclen != 0)) {
 				/* a multibyte character.  */
-				if (iswlower(wc))
-				{
+				if (iswlower(wc)) {
 					wc = towupper(wc);
-					wcrtomb(cp, wc, &prev_mbs);
+					wcrtomb((char *) cp, wc, &prev_mbs);
 				}
 				/* Adjust the pointer.  */
 				cp += mbclen - 1;
-		    } else {
+			} else {
 				/* Otherwise we treat it as a singlebyte character.  */
 				if (ISLOWER(*cp))
 					*cp = toupper(*cp);
-		    }
+			}
 		} else
 #endif
 		if (ISLOWER(*cp))
@@ -1602,9 +1698,9 @@ do_atan2(NODE *tree)
 	t1 = tree_eval(tree->lnode);
 	t2 = tree_eval(tree->rnode->lnode);
 	if (do_lint) {
-		if ((t1->flags & (NUM|NUMBER)) == 0)
+		if ((t1->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("atan2: received non-numeric first argument"));
-		if ((t2->flags & (NUM|NUMBER)) == 0)
+		if ((t2->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("atan2: received non-numeric second argument"));
 	}
 	d1 = force_number(t1);
@@ -1623,7 +1719,7 @@ do_sin(NODE *tree)
 	double d;
 
 	tmp = tree_eval(tree->lnode);
-	if (do_lint && (tmp->flags & (NUM|NUMBER)) == 0)
+	if (do_lint && (tmp->flags & (NUMCUR|NUMBER)) == 0)
 		lintwarn(_("sin: received non-numeric argument"));
 	d = sin((double) force_number(tmp));
 	free_temp(tmp);
@@ -1639,7 +1735,7 @@ do_cos(NODE *tree)
 	double d;
 
 	tmp = tree_eval(tree->lnode);
-	if (do_lint && (tmp->flags & (NUM|NUMBER)) == 0)
+	if (do_lint && (tmp->flags & (NUMCUR|NUMBER)) == 0)
 		lintwarn(_("cos: received non-numeric argument"));
 	d = cos((double) force_number(tmp));
 	free_temp(tmp);
@@ -1653,14 +1749,19 @@ static char state[512];
 
 /* ARGSUSED */
 NODE *
-do_rand(NODE *tree)
+do_rand(NODE *tree ATTRIBUTE_UNUSED)
 {
 	if (firstrand) {
 		(void) initstate((unsigned) 1, state, sizeof state);
 		srandom(1);
 		firstrand = FALSE;
 	}
-	return tmp_number((AWKNUM) random() / GAWK_RANDOM_MAX);
+	/*
+	 * Per historical practice and POSIX, return value N is
+	 *
+	 * 	0 <= n < 1
+	 */
+	return tmp_number((AWKNUM) (random() % GAWK_RANDOM_MAX) / GAWK_RANDOM_MAX);
 }
 
 /* do_srand --- seed the random number generator */
@@ -1683,7 +1784,7 @@ do_srand(NODE *tree)
 		srandom((unsigned int) (save_seed = (long) time((time_t *) 0)));
 	else {
 		tmp = tree_eval(tree->lnode);
-		if (do_lint && (tmp->flags & (NUM|NUMBER)) == 0)
+		if (do_lint && (tmp->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("srand: received non-numeric argument"));
 		srandom((unsigned int) (save_seed = (long) force_number(tmp)));
 		free_temp(tmp);
@@ -1693,7 +1794,7 @@ do_srand(NODE *tree)
 
 /* do_match --- match a regexp, set RSTART and RLENGTH,
  * 	optional third arg is array filled with text of
- * 	subpatterns enclosed in parens.
+ * 	subpatterns enclosed in parens and start and len info.
  */
 
 NODE *
@@ -1705,21 +1806,22 @@ do_match(NODE *tree)
 	Regexp *rp;
 	regoff_t s;
 	char *start;
+	char *buf = NULL;
+	char buff[100];
+	size_t amt, oldamt = 0, ilen, slen;
+	char *subsepstr;
+	size_t subseplen;
 
 	t1 = force_string(tree_eval(tree->lnode));
 	tree = tree->rnode;
 	rp = re_update(tree->lnode);
 
 	dest = NULL;
-	if (tree->rnode != NULL) {  /* 3rd optional arg for the subpatterns */
-		dest = tree->rnode->lnode;
-		if (dest->type == Node_param_list)
-			dest = stack_ptr[dest->param_cnt];
-		if (dest->type == Node_array_ref)
-			dest = dest->orig_array;
-		if (dest->type != Node_var && dest->type != Node_var_array)
+	if (tree->rnode != NULL) {	/* 3rd optional arg for the subpatterns */
+		dest = get_param(tree->rnode->lnode);
+		if (dest->type != Node_var_array)
 			fatal(_("match: third argument is not an array"));
-		dest->type = Node_var_array;
+
 		assoc_clear(dest);
 	}
 	
@@ -1730,14 +1832,55 @@ do_match(NODE *tree)
 	
 		/* Build the array only if the caller wants the optional subpatterns */
 		if (dest != NULL) {
-			for (ii = 0; (s = SUBPATSTART(rp, t1->stptr, ii)) != -1; ii++) {
-				start = t1->stptr + s;
-				len = SUBPATEND(rp, t1->stptr, ii) - s;
+			subsepstr = SUBSEP_node->var_value->stptr;
+			subseplen = SUBSEP_node->var_value->stlen;
 
-				it = make_string(start, len);
-				it->flags |= MAYBE_NUM;
-				*assoc_lookup(dest, tmp_number((AWKNUM) (ii)), FALSE) = it;
+			for (ii = 0; ii < NUMSUBPATS(rp, t1->stptr); ii++) {
+				/*
+				 * Loop over all the subpats; some of them may have
+				 * matched even if all of them did not.
+				 */
+				if ((s = SUBPATSTART(rp, t1->stptr, ii)) != -1) {
+					start = t1->stptr + s;
+					len = SUBPATEND(rp, t1->stptr, ii) - s;
+	
+					it = make_string(start, len);
+					/*
+					 * assoc_lookup() does free_temp() on 2nd arg.
+					 */
+					*assoc_lookup(dest, tmp_number((AWKNUM) (ii)), FALSE) = it;
+	
+					sprintf(buff, "%d", ii);
+					ilen = strlen(buff);
+					amt = ilen + subseplen + strlen("length") + 2;
+	
+					if (oldamt == 0) {
+						emalloc(buf, char *, amt, "do_match");
+					} else if (amt > oldamt) {
+						erealloc(buf, char *, amt, "do_match");
+					}
+					oldamt = amt;
+					memcpy(buf, buff, ilen);
+					memcpy(buf + ilen, subsepstr, subseplen);
+					memcpy(buf + ilen + subseplen, "start", 6);
+	
+					slen = ilen + subseplen + 5;
+	
+					it = make_number((AWKNUM) s + 1);
+					*assoc_lookup(dest, tmp_string(buf, slen), FALSE) = it;
+	
+					memcpy(buf, buff, ilen);
+					memcpy(buf + ilen, subsepstr, subseplen);
+					memcpy(buf + ilen + subseplen, "length", 7);
+	
+					slen = ilen + subseplen + 6;
+	
+					it = make_number((AWKNUM) len);
+					*assoc_lookup(dest, tmp_string(buf, slen), FALSE) = it;
+				}
 			}
+
+			free(buf);
 		}
 	} else {		/* match failed */
 		rstart = 0;
@@ -1817,7 +1960,7 @@ do_match(NODE *tree)
  */
 
 static NODE *
-sub_common(NODE *tree, int how_many, int backdigs)
+sub_common(NODE *tree, long how_many, int backdigs)
 {
 	register char *scan;
 	register char *bp, *cp;
@@ -1849,13 +1992,14 @@ sub_common(NODE *tree, int how_many, int backdigs)
 	char *mb_indices;
 #endif
 
-	tmp = tree->lnode;
+	tmp = tree->lnode;		/* regexp */
 	rp = re_update(tmp);
 
-	tree = tree->rnode;
+	tree = tree->rnode;		/* replacement text */
 	s = tree->lnode;
+	s = force_string(tree_eval(s));
 
-	tree = tree->rnode;
+	tree = tree->rnode;		/* original string */
 	tmp = tree->lnode;
 	t = force_string(tree_eval(tmp));
 
@@ -1863,6 +2007,7 @@ sub_common(NODE *tree, int how_many, int backdigs)
 	if (research(rp, t->stptr, 0, t->stlen, TRUE) == -1 ||
 	    RESTART(rp, t->stptr) > t->stlen) {
 		free_temp(t);
+		free_temp(s);
 		return tmp_number((AWKNUM) 0.0);
 	}
 
@@ -1883,7 +2028,6 @@ sub_common(NODE *tree, int how_many, int backdigs)
 	textlen = t->stlen;
 	buflen = textlen + 2;
 
-	s = force_string(tree_eval(s));
 	repl = s->stptr;
 	replend = repl + s->stlen;
 	repllen = replend - repl;
@@ -1900,7 +2044,7 @@ sub_common(NODE *tree, int how_many, int backdigs)
 	 * 	sub(/foo/, "", mystring)
 	 * for example.
 	 */
-	if (MB_CUR_MAX > 1 && repllen > 0) {
+	if (gawk_mb_cur_max > 1 && repllen > 0) {
 		emalloc(mb_indices, char *, repllen * sizeof(char), "sub_common");
 		index_multibyte_buffer(repl, mb_indices, repllen);
 	} else
@@ -1908,7 +2052,7 @@ sub_common(NODE *tree, int how_many, int backdigs)
 #endif
 	for (scan = repl; scan < replend; scan++) {
 #ifdef MBS_SUPPORT
-		if ((MB_CUR_MAX == 1 || (repllen > 0 && mb_indices[scan - repl] == 1))
+		if ((gawk_mb_cur_max == 1 || (repllen > 0 && mb_indices[scan - repl] == 1))
 			&& (*scan == '&')) {
 #else
 		if (*scan == '&') {
@@ -1969,9 +2113,13 @@ sub_common(NODE *tree, int how_many, int backdigs)
 			/*
 			 * If the current match matched the null string,
 			 * and the last match didn't and did a replacement,
-			 * then skip this one.
+			 * and the match of the null string is at the front of
+			 * the text (meaning right after end of the previous
+			 * replacement), then skip this one.
 			 */
-			if (lastmatchnonzero && matchstart == matchend) {
+			if (matchstart == matchend
+			    && lastmatchnonzero
+			    && matchstart == text) {
 				lastmatchnonzero = FALSE;
 				matches--;
 				goto empty;
@@ -1983,7 +2131,7 @@ sub_common(NODE *tree, int how_many, int backdigs)
 			 */
 			for (scan = repl; scan < replend; scan++)
 #ifdef MBS_SUPPORT
-				if ((MB_CUR_MAX == 1
+				if ((gawk_mb_cur_max == 1
 					 || (repllen > 0 && mb_indices[scan - repl] == 1))
 					&& (*scan == '&'))
 #else
@@ -1992,7 +2140,7 @@ sub_common(NODE *tree, int how_many, int backdigs)
 					for (cp = matchstart; cp < matchend; cp++)
 						*bp++ = *cp;
 #ifdef MBS_SUPPORT
-				else if ((MB_CUR_MAX == 1
+				else if ((gawk_mb_cur_max == 1
 					 || (repllen > 0 && mb_indices[scan - repl] == 1))
 						 && (*scan == '\\')) {
 #else
@@ -2081,7 +2229,7 @@ sub_common(NODE *tree, int how_many, int backdigs)
 		}
 		if (after_assign != NULL)
 			(*after_assign)();
-		t->flags &= ~(NUM|NUMBER);
+		t->flags &= ~(NUMCUR|NUMBER);
 	}
 #ifdef MBS_SUPPORT
 	if (mb_indices != NULL)
@@ -2142,19 +2290,21 @@ do_gensub(NODE *tree)
 	n3.lnode = target;
 	n2.rnode = & n3;
 
-	if ((t->flags & (STR|STRING)) != 0) {
+	if ((t->flags & (STRCUR|STRING)) != 0) {
 		if (t->stlen > 0 && (t->stptr[0] == 'g' || t->stptr[0] == 'G'))
 			how_many = -1;
 		else
 			how_many = 1;
 	} else {
 		d = force_number(t);
-		if (d > 0)
+		if (d < 1)
+			how_many = 1;
+		else if (d < LONG_MAX)
 			how_many = d;
 		else
-			how_many = 1;
+			how_many = LONG_MAX;
 		if (d == 0)
-			warning(_("gensub: 3rd argument of 0 treated as 1"));
+			warning(_("gensub: third argument of 0 treated as 1"));
 	}
 
 	free_temp(t);
@@ -2247,15 +2397,13 @@ sgfmt(char *buf,	/* return buffer; assumed big enough to hold result */
 }
 #endif	/* GFMT_WORKAROUND */
 
-#define BITS_PER_BYTE	8	/* if not true, you lose. too bad. */
-
 /* do_lshift --- perform a << operation */
 
 NODE *
 do_lshift(NODE *tree)
 {
 	NODE *s1, *s2;
-	unsigned long uval, ushift, res;
+	uintmax_t uval, ushift, res;
 	AWKNUM val, shift;
 
 	s1 = tree_eval(tree->lnode);
@@ -2264,23 +2412,23 @@ do_lshift(NODE *tree)
 	shift = force_number(s2);
 
 	if (do_lint) {
-		if ((s1->flags & (NUM|NUMBER)) == 0)
+		if ((s1->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("lshift: received non-numeric first argument"));
-		if ((s2->flags & (NUM|NUMBER)) == 0)
+		if ((s2->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("lshift: received non-numeric first argument"));
 		if (val < 0 || shift < 0)
 			lintwarn(_("lshift(%lf, %lf): negative values will give strange results"), val, shift);
 		if (double_to_int(val) != val || double_to_int(shift) != shift)
 			lintwarn(_("lshift(%lf, %lf): fractional values will be truncated"), val, shift);
-		if (shift > (sizeof(unsigned long) * BITS_PER_BYTE))
+		if (shift >= sizeof(uintmax_t) * CHAR_BIT)
 			lintwarn(_("lshift(%lf, %lf): too large shift value will give strange results"), val, shift);
 	}
 
 	free_temp(s1);
 	free_temp(s2);
 
-	uval = (unsigned long) val;
-	ushift = (unsigned long) shift;
+	uval = (uintmax_t) val;
+	ushift = (uintmax_t) shift;
 
 	res = uval << ushift;
 	return tmp_number((AWKNUM) res);
@@ -2292,7 +2440,7 @@ NODE *
 do_rshift(NODE *tree)
 {
 	NODE *s1, *s2;
-	unsigned long uval, ushift, res;
+	uintmax_t uval, ushift, res;
 	AWKNUM val, shift;
 
 	s1 = tree_eval(tree->lnode);
@@ -2301,23 +2449,23 @@ do_rshift(NODE *tree)
 	shift = force_number(s2);
 
 	if (do_lint) {
-		if ((s1->flags & (NUM|NUMBER)) == 0)
+		if ((s1->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("rshift: received non-numeric first argument"));
-		if ((s2->flags & (NUM|NUMBER)) == 0)
+		if ((s2->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("rshift: received non-numeric first argument"));
 		if (val < 0 || shift < 0)
 			lintwarn(_("rshift(%lf, %lf): negative values will give strange results"), val, shift);
 		if (double_to_int(val) != val || double_to_int(shift) != shift)
 			lintwarn(_("rshift(%lf, %lf): fractional values will be truncated"), val, shift);
-		if (shift > (sizeof(unsigned long) * BITS_PER_BYTE))
+		if (shift >= sizeof(uintmax_t) * CHAR_BIT)
 			lintwarn(_("rshift(%lf, %lf): too large shift value will give strange results"), val, shift);
 	}
 
 	free_temp(s1);
 	free_temp(s2);
 
-	uval = (unsigned long) val;
-	ushift = (unsigned long) shift;
+	uval = (uintmax_t) val;
+	ushift = (uintmax_t) shift;
 
 	res = uval >> ushift;
 	return tmp_number((AWKNUM) res);
@@ -2329,7 +2477,7 @@ NODE *
 do_and(NODE *tree)
 {
 	NODE *s1, *s2;
-	unsigned long uleft, uright, res;
+	uintmax_t uleft, uright, res;
 	AWKNUM left, right;
 
 	s1 = tree_eval(tree->lnode);
@@ -2338,9 +2486,9 @@ do_and(NODE *tree)
 	right = force_number(s2);
 
 	if (do_lint) {
-		if ((s1->flags & (NUM|NUMBER)) == 0)
+		if ((s1->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("and: received non-numeric first argument"));
-		if ((s2->flags & (NUM|NUMBER)) == 0)
+		if ((s2->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("and: received non-numeric first argument"));
 		if (left < 0 || right < 0)
 			lintwarn(_("and(%lf, %lf): negative values will give strange results"), left, right);
@@ -2351,8 +2499,8 @@ do_and(NODE *tree)
 	free_temp(s1);
 	free_temp(s2);
 
-	uleft = (unsigned long) left;
-	uright = (unsigned long) right;
+	uleft = (uintmax_t) left;
+	uright = (uintmax_t) right;
 
 	res = uleft & uright;
 	return tmp_number((AWKNUM) res);
@@ -2364,7 +2512,7 @@ NODE *
 do_or(NODE *tree)
 {
 	NODE *s1, *s2;
-	unsigned long uleft, uright, res;
+	uintmax_t uleft, uright, res;
 	AWKNUM left, right;
 
 	s1 = tree_eval(tree->lnode);
@@ -2373,9 +2521,9 @@ do_or(NODE *tree)
 	right = force_number(s2);
 
 	if (do_lint) {
-		if ((s1->flags & (NUM|NUMBER)) == 0)
+		if ((s1->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("or: received non-numeric first argument"));
-		if ((s2->flags & (NUM|NUMBER)) == 0)
+		if ((s2->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("or: received non-numeric first argument"));
 		if (left < 0 || right < 0)
 			lintwarn(_("or(%lf, %lf): negative values will give strange results"), left, right);
@@ -2386,8 +2534,8 @@ do_or(NODE *tree)
 	free_temp(s1);
 	free_temp(s2);
 
-	uleft = (unsigned long) left;
-	uright = (unsigned long) right;
+	uleft = (uintmax_t) left;
+	uright = (uintmax_t) right;
 
 	res = uleft | uright;
 	return tmp_number((AWKNUM) res);
@@ -2399,7 +2547,7 @@ NODE *
 do_xor(NODE *tree)
 {
 	NODE *s1, *s2;
-	unsigned long uleft, uright, res;
+	uintmax_t uleft, uright, res;
 	AWKNUM left, right;
 
 	s1 = tree_eval(tree->lnode);
@@ -2408,9 +2556,9 @@ do_xor(NODE *tree)
 	right = force_number(s2);
 
 	if (do_lint) {
-		if ((s1->flags & (NUM|NUMBER)) == 0)
+		if ((s1->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("xor: received non-numeric first argument"));
-		if ((s2->flags & (NUM|NUMBER)) == 0)
+		if ((s2->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("xor: received non-numeric first argument"));
 		if (left < 0 || right < 0)
 			lintwarn(_("xor(%lf, %lf): negative values will give strange results"), left, right);
@@ -2421,8 +2569,8 @@ do_xor(NODE *tree)
 	free_temp(s1);
 	free_temp(s2);
 
-	uleft = (unsigned long) left;
-	uright = (unsigned long) right;
+	uleft = (uintmax_t) left;
+	uright = (uintmax_t) right;
 
 	res = uleft ^ uright;
 	return tmp_number((AWKNUM) res);
@@ -2435,14 +2583,14 @@ do_compl(NODE *tree)
 {
 	NODE *tmp;
 	double d;
-	unsigned long uval;
+	uintmax_t uval;
 
 	tmp = tree_eval(tree->lnode);
 	d = force_number(tmp);
 	free_temp(tmp);
 
 	if (do_lint) {
-		if ((tmp->flags & (NUM|NUMBER)) == 0)
+		if ((tmp->flags & (NUMCUR|NUMBER)) == 0)
 			lintwarn(_("compl: received non-numeric argument"));
 		if (d < 0)
 			lintwarn(_("compl(%lf): negative value will give strange results"), d);
@@ -2450,7 +2598,7 @@ do_compl(NODE *tree)
 			lintwarn(_("compl(%lf): fractional value will be truncated"), d);
 	}
 
-	uval = (unsigned long) d;
+	uval = (uintmax_t) d;
 	uval = ~ uval;
 	return tmp_number((AWKNUM) uval);
 }
@@ -2465,7 +2613,9 @@ do_strtonum(NODE *tree)
 
 	tmp = tree_eval(tree->lnode);
 
-	if (isnondecimal(tmp->stptr))
+	if ((tmp->flags & (NUMBER|NUMCUR)) != 0)
+		d = (double) force_number(tmp);
+	else if (isnondecimal(tmp->stptr))
 		d = nondec2awknum(tmp->stptr, tmp->stlen);
 	else
 		d = (double) force_number(tmp);
@@ -2561,9 +2711,9 @@ done:
 static int
 localecategory_from_argument(NODE *tree)
 {
-	static struct category_table {
+	static const struct category_table {
 		int val;
-		char *name;
+		const char *name;
 	} cat_tab[] = {
 #ifdef LC_ALL
 		{ LC_ALL,	"LC_ALL" },
@@ -2683,7 +2833,7 @@ do_dcngettext(NODE *tree)
 {
 	NODE *tmp, *t1, *t2, *t3;
 	char *string1, *string2;
-	long number;
+	unsigned long number;
 	char *the_result;
 #if ENABLE_NLS && HAVE_LC_MESSAGES && HAVE_DCGETTEXT
 	int lc_cat;
@@ -2699,7 +2849,7 @@ do_dcngettext(NODE *tree)
 	string2 = t2->stptr;
 
 	tmp = tree->rnode->rnode->lnode;	/* third argument */
-	number = (long) double_to_int(force_number(tree_eval(tmp)));
+	number = (unsigned long) double_to_int(force_number(tree_eval(tmp)));
 
 	t3 = NULL;
 #if ENABLE_NLS && HAVE_LC_MESSAGES && HAVE_DCGETTEXT
